@@ -6,13 +6,14 @@ except ImportError:
 import sys
 
 from .utils import InheritedStuff
-from .utils import Stuff
+from .utils import Stuff, DummyModel
 
 from functools import partial
 from transitions import Machine, MachineError, State, EventData
 from transitions.core import listify, _prep_ordered_arg
 from unittest import TestCase, skipIf
 import warnings
+import weakref
 
 try:
     from unittest.mock import MagicMock
@@ -63,6 +64,20 @@ class TestTransitions(TestCase):
         self.assertEqual(listify((4, 5)), (4, 5))
         self.assertEqual(listify([1, 3]), [1, 3])
 
+        class Foo:
+            pass
+        obj = Foo()
+        proxy = weakref.proxy(obj)
+        del obj
+        self.assertEqual(listify(proxy), [proxy])
+
+    def test_weakproxy_model(self):
+        d = DummyModel()
+        pr = weakref.proxy(d)
+        self.machine_cls(pr, states=['A', 'B'], transitions=[['go', 'A', 'B']], initial='A')
+        pr.go()
+        self.assertTrue(pr.is_B())
+
     def test_property_initial(self):
         states = ['A', 'B', 'C', 'D']
         # Define with list of dictionaries
@@ -607,11 +622,9 @@ class TestTransitions(TestCase):
         def after_change(machine):
             machine.to_C(machine)
 
-        def failed_transition(machine):
-            raise ValueError('Something was wrong')
-
         states = ['A', 'B', 'C']
-        transitions = [{'trigger': 'do', 'source': '*', 'dest': 'C', 'before': failed_transition}]
+        transitions = [{'trigger': 'do', 'source': '*', 'dest': 'C',
+                        'before': partial(self.stuff.this_raises, ValueError)}]
         m = Machine(states=states, transitions=transitions, queued=True,
                     before_state_change=before_change, after_state_change=after_change)
         with self.assertRaises(MachineError):
@@ -620,6 +633,46 @@ class TestTransitions(TestCase):
         with self.assertRaises(ValueError):
             m.do(machine=m)
 
+    def test_queued_remove(self):
+        m = self.machine_cls(model=None, states=['A', 'B', 'C'], initial='A', queued=True)
+        assert_equal = self.assertEqual
+
+        class BaseModel:
+            def on_enter_A(self):
+                pass
+
+            def on_enter_B(self):
+                pass
+
+            def on_enter_C(self):
+                pass
+
+        class SubModel(BaseModel):
+            def __init__(self):
+                self.inner = BaseModel()
+
+            def on_enter_A(self):
+                self.to_B()
+                self.inner.to_B()
+
+            def on_enter_B(self):
+                self.to_C()
+                self.inner.to_C()
+                # queue should contain to_B(), inner.to_B(), to_C(), inner.to_C()
+                assert_equal(4, len(m._transition_queue))
+                m.remove_model(self)
+                # since to_B() is currently executed it should still be in the list, to_C should be gone
+                assert_equal(3, len(m._transition_queue))
+
+            def on_enter_C(self):
+                raise RuntimeError("Event was not cancelled")
+        model = SubModel()
+        m.add_model([model, model.inner])
+        model.to_A()
+        # test whether models can be removed outside event queue
+        m.remove_model(model.inner)
+        self.assertTrue(model.inner.is_C())
+
     def test___getattr___and_identify_callback(self):
         m = self.machine_cls(Stuff(), states=['A', 'B', 'C'], initial='A')
         m.add_transition('move', 'A', 'B')
@@ -782,6 +835,12 @@ class TestTransitions(TestCase):
         self.assertEqual(s1.state, 'B')
         self.assertEqual(s2.state, 'C')
 
+    def test_remove_model(self):
+        m = self.machine_cls()
+        self.assertIn(m, m.models)
+        m.remove_model(m)
+        self.assertNotIn(m, m.models)
+
     def test_string_trigger(self):
         def return_value(value):
             return value
@@ -799,17 +858,16 @@ class TestTransitions(TestCase):
         self.assertTrue(self.stuff.is_A())
         with self.assertRaises(AttributeError):
             self.stuff.trigger('not_available')
+        with self.assertRaises(MachineError):
+            self.stuff.trigger('maybe')
 
         model = Model()
         m = Machine(model=model)
         self.assertEqual(model.trigger(5), 5)
-
-        def raise_key_error():
-            raise KeyError
-
-        self.stuff.machine.add_transition('do_raises_keyerror', '*', 'C', before=raise_key_error)
+        self.stuff.machine.add_transition('do_raise_keyerror', '*', 'C',
+                                          before=partial(self.stuff.this_raises, KeyError))
         with self.assertRaises(KeyError):
-            self.stuff.trigger('do_raises_keyerror')
+            self.stuff.trigger('do_raise_keyerror')
 
         self.stuff.machine.get_model_state(self.stuff).ignore_invalid_triggers = True
         self.stuff.trigger('should_not_raise_anything')
@@ -828,7 +886,11 @@ class TestTransitions(TestCase):
         self.assertEqual(len(machine.get_triggers('B')), 0)
         self.assertEqual(len(machine.get_triggers('C')), 1)
         # self stuff machine should have to-transitions to every state
-        self.assertEqual(len(self.stuff.machine.get_triggers('B')), len(self.stuff.machine.states))
+        m = self.stuff.machine
+        self.assertEqual(len(m.get_triggers('B')), len(m.states))
+        trigger_name = m.get_triggers('B')
+        trigger_state = m.get_triggers(m.states['B'])
+        self.assertEqual(trigger_name, trigger_state)
 
     def test_skip_override(self):
         local_mock = MagicMock()
@@ -918,13 +980,11 @@ class TestTransitions(TestCase):
         def always_fails(event_data):
             return False
 
-        def always_raises(event_data):
-            raise Exception()
-
         transitions = [
             {'trigger': 'go', 'source': 'A', 'dest': 'B'},
             {'trigger': 'planA', 'source': 'B', 'dest': 'A', 'conditions': always_fails},
-            {'trigger': 'planB', 'source': 'B', 'dest': 'A', 'conditions': always_raises}
+            {'trigger': 'planB', 'source': 'B', 'dest': 'A',
+             'conditions': partial(self.stuff.this_raises, RuntimeError)}
         ]
         m = self.stuff.machine_cls(states=['A', 'B'], transitions=transitions,
                                    finalize_event=finalize_mock, initial='A', send_event=True)
@@ -937,22 +997,23 @@ class TestTransitions(TestCase):
         self.assertIsInstance(event_data, EventData)
         self.assertEqual(finalize_mock.call_count, 2)
         self.assertFalse(event_data.result)
-        with self.assertRaises(Exception):
+        with self.assertRaises(RuntimeError):
             m.planB()
-        self.assertEqual(finalize_mock.call_count, 3)
+
+        m.finalize_event.append(partial(self.stuff.this_raises, ValueError))
+        # ValueError in finalize should be suppressed
+        # but mock should have been called anyway
+        with self.assertRaises(RuntimeError):
+            m.planB()
+        self.assertEqual(finalize_mock.call_count, 4)
 
     def test_machine_finalize_exception(self):
 
-        exception = ZeroDivisionError()
-
-        def always_raises(event):
-            raise exception
-
         def finalize_callback(event):
-            self.assertEqual(event.error, exception)
+            self.assertIsInstance(event.error, ZeroDivisionError)
 
         m = self.stuff.machine_cls(states=['A', 'B'], send_event=True, initial='A',
-                                   before_state_change=always_raises,
+                                   before_state_change=partial(self.stuff.this_raises, ZeroDivisionError),
                                    finalize_event=finalize_callback)
 
         with self.assertRaises(ZeroDivisionError):
@@ -1002,7 +1063,7 @@ class TestTransitions(TestCase):
 
     def test_get_transitions(self):
         states = ['A', 'B', 'C', 'D']
-        m = Machine('self', states, initial='a', auto_transitions=False)
+        m = self.machine_cls('self', states, initial='A', auto_transitions=False)
         m.add_transition('go', ['A', 'B', 'C'], 'D')
         m.add_transition('run', 'A', 'D')
         self.assertEqual(
@@ -1012,6 +1073,18 @@ class TestTransitions(TestCase):
             [(t.source, t.dest)
              for t in m.get_transitions(source='A', dest='D')],
             [('A', 'D'), ('A', 'D')])
+        self.assertEqual(
+            sorted([(t.source, t.dest)
+                    for t in m.get_transitions(dest='D')]),
+            [('A', 'D'), ('A', 'D'), ('B', 'D'), ('C', 'D')])
+        self.assertEqual(
+            [(t.source, t.dest)
+             for t in m.get_transitions(source=m.states['A'], dest=m.states['D'])],
+            [('A', 'D'), ('A', 'D')])
+        self.assertEqual(
+            sorted([(t.source, t.dest)
+                    for t in m.get_transitions(dest=m.states['D'])]),
+            [('A', 'D'), ('A', 'D'), ('B', 'D'), ('C', 'D')])
 
     def test_remove_transition(self):
         self.stuff.machine.add_transition('go', ['A', 'B', 'C'], 'D')
@@ -1059,7 +1132,7 @@ class TestTransitions(TestCase):
                 self.status = None
                 self.state = 'some_value'
 
-        m = Machine(Model(), states=['A', 'B'], initial='A', model_attribute='status')
+        m = self.machine_cls(Model(), states=['A', 'B'], initial='A', model_attribute='status')
         self.assertEqual(m.model.status, 'A')
         self.assertEqual(m.model.state, 'some_value')
 
@@ -1111,6 +1184,34 @@ class TestTransitions(TestCase):
         with self.assertRaises(ValueError):
             m.add_transition(m.model_attribute, "A", "B")
 
+    def test_new_state_in_enter_callback(self):
+
+        machine = self.machine_cls(states=['A', 'B'], initial='A')
+
+        def on_enter_B():
+            state = self.machine_cls.state_cls(name='C')
+            machine.add_state(state)
+            machine.to_C()
+
+        machine.on_enter_B(on_enter_B)
+        machine.to_B()
+
+    def test_on_exception_callback(self):
+        mock = MagicMock()
+
+        def on_exception(event_data):
+            self.assertIsInstance(event_data.error, ValueError)
+            mock()
+
+        m = self.machine_cls(states=['A', 'B'], initial='A', send_event=True,
+                             after_state_change=partial(self.stuff.this_raises, ValueError))
+        with self.assertRaises(ValueError):
+            m.to_B()
+
+        m.on_exception.append(on_exception)
+        m.to_B()
+        self.assertTrue(mock.called)
+
 
 class TestWarnings(TestCase):
     def test_multiple_machines_per_model(self):
