@@ -425,6 +425,33 @@ class TestNestedTransitions(TestTransitions):
         self.assertEqual(len(trans), 3)
         self.assertTrue('relax' in trans)
 
+    def test_get_nested_transitions(self):
+        seperator = self.state_cls.separator
+        states = ['A', {'name': 'B', 'children': ['1', '2', {'name': '3', 'children': ['a', 'b'],
+                                                             'transitions': [['inner', 'a', 'b'],
+                                                                             ['inner', 'b', 'a']]}],
+                        'transitions': [['mid', '1', '1'],
+                                        ['mid', '2', '3'],
+                                        ['mid', '3', '1'],
+                                        ['mid2', '2', '3'],
+                                        ['mid_loop', '1', '1']]}]
+        transitions = [['outer', 'A', 'B'],
+                       ['outer', ['A', 'B'], 'C']]
+        machine = self.stuff.machine_cls(states=states, transitions=transitions, initial='A', auto_transitions=False)
+        self.assertEqual(10, len(machine.get_transitions()))
+        self.assertEqual(2, len(machine.get_transitions(source='A')))
+        self.assertEqual(2, len(machine.get_transitions('inner')))
+        self.assertEqual(3, len(machine.get_transitions('mid')))
+        self.assertEqual(3, len(machine.get_transitions(dest='B{0}1'.format(seperator))))
+        self.assertEqual(2, len(machine.get_transitions(source='B{0}2'.format(seperator),
+                                                        dest='B{0}3'.format(seperator))))
+        self.assertEqual(1, len(machine.get_transitions(source='B{0}3{0}a'.format(seperator),
+                                                        dest='B{0}3{0}b'.format(seperator))))
+        self.assertEqual(1, len(machine.get_transitions(source=machine.states['B'].states['3'].states['b'])))
+        # should be B_3_b -> B_3_a, B_3 -> B_1 and B -> C
+        self.assertEqual(3, len(machine.get_transitions(source=machine.states['B'].states['3'].states['b'],
+                                                        delegate=True)))
+
     def test_internal_transitions(self):
         s = self.stuff
         s.machine.add_transition('internal', 'A', None, prepare='increase_level')
@@ -514,6 +541,61 @@ class TestNestedTransitions(TestTransitions):
                                    transitions=[['go', 'B', 'A']], initial='A')
         self.assertFalse(False, machine.go())
 
+    def test_exception_in_state_enter_exit(self):
+        # https://github.com/pytransitions/transitions/issues/486
+        # NestedState._scope needs to be reset when an error is raised in a state callback
+        class Model:
+            def on_enter_B_1(self):
+                raise RuntimeError("Oh no!")
+
+            def on_exit_C_1(self):
+                raise RuntimeError("Oh no!")
+
+        states = ['A',
+                  {'name': 'B', 'initial': '1', 'children': ['1', '2']},
+                  {'name': 'C', 'initial': '1', 'children': ['1', '2']}]
+        model = Model()
+        machine = self.machine_cls(model, states=states, initial='A')
+        with self.assertRaises(RuntimeError):
+            model.to_B()
+        self.assertTrue(model.is_B_1())
+        machine.set_state('A', model)
+        with self.assertRaises(RuntimeError):
+            model.to_B()
+        with self.assertRaises(RuntimeError):
+            model.to_C()
+            model.to_A()
+        self.assertTrue(model.is_C_1())
+        machine.set_state('A', model)
+        model.to_C()
+        self.assertTrue(model.is_C_1())
+
+    def test_correct_subclassing(self):
+        from transitions.core import State
+
+        class WrongStateClass(self.machine_cls):
+            state_cls = State
+
+        class MyNestedState(NestedState):
+            pass
+
+        class CorrectStateClass(self.machine_cls):
+            state_cls = MyNestedState
+
+        with self.assertRaises(AssertionError):
+            m = WrongStateClass()
+        m = CorrectStateClass()
+
+    def test_queued_callbacks(self):
+        states = [
+            "initial",
+            {'name': 'A', 'children': [{'name': '1', 'on_enter': 'go'}, '2'],
+             'transitions': [['go', '1', '2']], 'initial': '1'}
+        ]
+        machine = self.machine_cls(states=states, initial='initial', queued=True)
+        machine.to_A()
+        self.assertEqual("A{0}2".format(self.state_cls.separator), machine.state)
+
 
 class TestSeparatorsBase(TestCase):
 
@@ -671,8 +753,18 @@ class TestSeparatorsBase(TestCase):
         machine.to_C()  # exit B, enter C
         if separator == '_':
             machine.to_C_3_a()
+            self.assertTrue(machine.is_C(allow_substates=True))
+            self.assertFalse(machine.is_C())
+            self.assertTrue(machine.is_C_3(allow_substates=True))
+            self.assertFalse(machine.is_C_3())
+            self.assertTrue(machine.is_C_3_a())
         else:
             machine.to_C.s3.a()  # enter C↦a; enter C↦3↦a;
+            self.assertTrue(machine.is_C(allow_substates=True))
+            self.assertFalse(machine.is_C())
+            self.assertTrue(machine.is_C.s3(allow_substates=True))
+            self.assertFalse(machine.is_C.s3())
+            self.assertTrue(machine.is_C.s3.a())
         self.assertEqual(machine.state, 'C{0}3{0}a'.format(separator))
         machine.to('C{0}2'.format(separator))  # exit C↦3↦a, exit C↦3, enter C↦2
         self.assertEqual(machine.state, 'C{0}2'.format(separator))
