@@ -50,7 +50,11 @@ def listify(obj):
     if obj is None:
         return []
 
-    return obj if isinstance(obj, (list, tuple, EnumMeta)) else [obj]
+    try:
+        return obj if isinstance(obj, (list, tuple, EnumMeta)) else [obj]
+    except ReferenceError:
+        # obj is an empty weakref
+        return [obj]
 
 
 def _prep_ordered_arg(desired_length, arguments=None):
@@ -376,9 +380,13 @@ class Event(object):
         self.transitions[transition.source].append(transition)
 
     def trigger(self, model, *args, **kwargs):
-        """ Serially execute all transitions that match the current state,
-        halting as soon as one successfully completes.
+        """ Executes all transitions that match the current state,
+        halting as soon as one successfully completes. More precisely, it prepares a partial
+        of the internal ``_trigger`` function, passes this to ``Machine._process``.
+        It is up to the machine's configuration of the Event whether processing happens queued (sequentially) or
+        whether further Events are processed as they occur.
         Args:
+            model (object): The currently processed model
             args and kwargs: Optional positional or named arguments that will
                 be passed onto the EventData object, enabling arbitrary state
                 information to be passed on to downstream triggered functions.
@@ -394,7 +402,14 @@ class Event(object):
 
     def _trigger(self, model, *args, **kwargs):
         """ Internal trigger function called by the ``Machine`` instance. This should not
-        be called directly but via the public method ``Machine.trigger``.
+        be called directly but via the public method ``Machine.process``.
+        Args:
+            model (object): The currently processed model
+            args and kwargs: Optional positional or named arguments that will
+                be passed onto the EventData object, enabling arbitrary state
+                information to be passed on to downstream triggered functions.
+        Returns: boolean indicating whether or not a transition was
+            successfully executed (True if successful, False if not).
         """
         state = self.machine.get_model_state(model)
         if state.name not in self.transitions:
@@ -422,10 +437,19 @@ class Event(object):
                     break
         except Exception as err:
             event_data.error = err
-            raise
+            if self.machine.on_exception:
+                self.machine.callbacks(self.machine.on_exception, event_data)
+            else:
+                raise
         finally:
-            self.machine.callbacks(self.machine.finalize_event, event_data)
-            _LOGGER.debug("%sExecuted machine finalize callbacks", self.machine.name)
+            try:
+                self.machine.callbacks(self.machine.finalize_event, event_data)
+                _LOGGER.debug("%sExecuted machine finalize callbacks", self.machine.name)
+            except Exception as err:
+                _LOGGER.error("%sWhile executing finalize callbacks a %s occurred: %s.",
+                              self.machine.name,
+                              type(err).__name__,
+                              str(err))
 
         return event_data.result
 
@@ -483,7 +507,8 @@ class Machine(object):
                  send_event=False, auto_transitions=True,
                  ordered_transitions=False, ignore_invalid_triggers=None,
                  before_state_change=None, after_state_change=None, name=None,
-                 queued=False, prepare_event=None, finalize_event=None, model_attribute='state', **kwargs):
+                 queued=False, prepare_event=None, finalize_event=None, model_attribute='state', on_exception=None,
+                 **kwargs):
         """
         Args:
             model (object or list): The object(s) whose states we want to manage. If 'self',
@@ -527,6 +552,8 @@ class Machine(object):
                 It receives the very same args as normal callbacks.
             finalize_event: A callable called on for each triggered event after transitions have been processed.
                 This is also called when a transition raises an exception.
+            on_exception: A callable called when an event raises an exception. If not set,
+                the exception will be raised instead.
 
             **kwargs additional arguments passed to next class in MRO. This can be ignored in most cases.
         """
@@ -545,6 +572,7 @@ class Machine(object):
         self._after_state_change = []
         self._prepare_event = []
         self._finalize_event = []
+        self._on_exception = []
         self._initial = None
 
         self.states = OrderedDict()
@@ -556,6 +584,7 @@ class Machine(object):
         self.before_state_change = before_state_change
         self.after_state_change = after_state_change
         self.finalize_event = finalize_event
+        self.on_exception = on_exception
         self.name = name + ": " if name is not None else ""
         self.model_attribute = model_attribute
 
@@ -602,11 +631,16 @@ class Machine(object):
 
     def remove_model(self, model):
         """ Remove a model from the state machine. The model will still contain all previously added triggers
-        and callbacks, but will not receive updates when states or transitions are added to the Machine. """
+        and callbacks, but will not receive updates when states or transitions are added to the Machine.
+        If an event queue is used, all queued events of that model will be removed."""
         models = listify(model)
 
         for mod in models:
             self.models.remove(mod)
+        if len(self._transition_queue) > 0:
+            # the first element of the list is currently executed. Keeping it for further Machine._process(ing)
+            self._transition_queue = deque(
+                [self._transition_queue[0]] + [e for e in self._transition_queue if e.args[0] not in models])
 
     @classmethod
     def _create_transition(cls, *args, **kwargs):
@@ -697,6 +731,16 @@ class Machine(object):
     def finalize_event(self, value):
         self._finalize_event = listify(value)
 
+    @property
+    def on_exception(self):
+        """Callbacks will be executed when an Event raises an Exception."""
+        return self._on_exception
+
+    # this should make sure that finalize_event is always a list
+    @on_exception.setter
+    def on_exception(self, value):
+        self._on_exception = listify(value)
+
     def get_state(self, state):
         """ Return the State instance with the passed name. """
         if isinstance(state, Enum):
@@ -836,7 +880,6 @@ class Machine(object):
         """Convenience function added to the model to trigger events by name.
         Args:
             model (object): Model with assigned event trigger.
-            machine (Machine): The machine containing the evaluated events.
             trigger_name (str): Name of the trigger to be called.
             *args: Variable length argument list which is passed to the triggered event.
             **kwargs: Arbitrary keyword arguments which is passed to the triggered event.
@@ -862,8 +905,8 @@ class Machine(object):
         Returns:
             list of transition/trigger names.
         """
-        states = set(args)
-        return [t for (t, ev) in self.events.items() if any(state in ev.transitions for state in states)]
+        names = set([state.name if hasattr(state, 'name') else state for state in args])
+        return [t for (t, ev) in self.events.items() if any(name in ev.transitions for name in names)]
 
     def add_transition(self, trigger, source, dest, conditions=None,
                        unless=None, before=None, after=None, prepare=None, **kwargs):
@@ -1014,8 +1057,8 @@ class Machine(object):
         """ Return the transitions from the Machine.
         Args:
             trigger (str): Trigger name of the transition.
-            source (str): Limits list to transitions from a certain state.
-            dest (str): Limits list to transitions to a certain state.
+            source (str, Enum or State): Limits list to transitions from a certain state.
+            dest (str, Enum or State): Limits list to transitions to a certain state.
         """
         if trigger:
             try:
@@ -1028,11 +1071,12 @@ class Machine(object):
         for event in events:
             transitions.extend(
                 itertools.chain.from_iterable(event.transitions.values()))
+        target_source = source.name if hasattr(source, 'name') else source if source != "*" else ""
+        target_dest = dest.name if hasattr(dest, 'name') else dest if dest != "*" else ""
         return [transition
                 for transition in transitions
-                if (transition.source, transition.dest) == (
-                    source if source != "*" else transition.source,
-                    dest if dest != "*" else transition.dest)]
+                if (transition.source, transition.dest) == (target_source or transition.source,
+                                                            target_dest or transition.dest)]
 
     def remove_transition(self, trigger, source="*", dest="*"):
         """ Removes a transition from the Machine and all models.
