@@ -3,7 +3,7 @@ import logging
 import asyncio
 import contextvars
 import inspect
-
+from collections import defaultdict, deque
 from functools import partial, reduce
 import copy
 
@@ -120,7 +120,7 @@ class AsyncTransition(Transition):
 
         machine = event_data.machine
         # cancel running tasks since the transition will happen
-        machine.switch_model_context(event_data.model)
+        await machine.switch_model_context(event_data.model)
 
         await event_data.machine.callbacks(event_data.machine.before_state_change, event_data)
         await event_data.machine.callbacks(self.before, event_data)
@@ -136,12 +136,12 @@ class AsyncTransition(Transition):
 
     async def _change_state(self, event_data):
         if hasattr(event_data.machine, "model_graphs"):
-            graph = event_data.machine.model_graphs[event_data.model]
+            graph = event_data.machine.model_graphs[id(event_data.model)]
             graph.reset_styling()
             graph.set_previous_transition(self.source, self.dest)
         await event_data.machine.get_state(self.source).exit(event_data)
         event_data.machine.set_state(self.dest, event_data.model)
-        event_data.update(event_data.model.state)
+        event_data.update(getattr(event_data.model, event_data.machine.model_attribute))
         await event_data.machine.get_state(self.dest).enter(event_data)
 
 
@@ -149,7 +149,7 @@ class NestedAsyncTransition(AsyncTransition, NestedTransition):
 
     async def _change_state(self, event_data):
         if hasattr(event_data.machine, "model_graphs"):
-            graph = event_data.machine.model_graphs[event_data.model]
+            graph = event_data.machine.model_graphs[id(event_data.model)]
             graph.reset_styling()
             graph.set_previous_transition(self.source, self.dest)
         state_tree, exit_partials, enter_partials = self._resolve_transition(event_data)
@@ -177,7 +177,7 @@ class AsyncEvent(Event):
         return await self.machine.process_context(func, _model)
 
     async def _trigger(self, model, *args, **kwargs):
-        state = self.machine.get_state(model.state)
+        state = self.machine.get_state(getattr(model, self.machine.model_attribute))
         if state.name not in self.transitions:
             msg = "%sCan't trigger event %s from state %s!" % (self.machine.name, self.name,
                                                                state.name)
@@ -203,7 +203,10 @@ class AsyncEvent(Event):
                     break
         except Exception as err:
             event_data.error = err
-            raise
+            if self.machine.on_exception:
+                await self.machine.callbacks(self.machine.on_exception, event_data)
+            else:
+                raise
         finally:
             await self.machine.callbacks(self.machine.finalize_event, event_data)
             _LOGGER.debug("%sExecuted machine finalize callbacks", self.machine.name)
@@ -263,7 +266,10 @@ class NestedAsyncEvent(NestedEvent):
                     break
         except Exception as err:
             event_data.error = err
-            raise
+            if self.machine.on_exception:
+                await self.machine.callbacks(self.machine.on_exception, event_data)
+            else:
+                raise
         finally:
             await machine.callbacks(machine.finalize_event, event_data)
             _LOGGER.debug("%sExecuted machine finalize callbacks", machine.name)
@@ -287,6 +293,8 @@ class AsyncMachine(Machine):
             Callbacks will be executed AFTER the custom callbacks assigned to the transition.
         finalize_event (list): Callbacks will be executed after all transitions callbacks have been executed.
             Callbacks mentioned here will also be called if a transition or condition check raised an error.
+        on_exception: A callable called when an event raises an exception. If not set,
+            the Exception will be raised instead.
         queued (bool): Whether transitions in callbacks should be executed immediately (False) or sequentially.
         send_event (bool): When True, any arguments passed to trigger methods will be wrapped in an EventData
             object, allowing indirect and encapsulated access to data. When False, all positional and keyword
@@ -306,6 +314,20 @@ class AsyncMachine(Machine):
     protected_tasks = []
     current_context = contextvars.ContextVar('current_context', default=None)
 
+    def __init__(self, *args, **kwargs):
+        self._transition_queue_dict = {}
+        super().__init__(*args, **kwargs)
+        if self.has_queue is True:
+            self._transition_queue_dict = _DictionaryMock()
+            # _DictionaryMock sets and returns ONE internal value and ignores the passed key
+            self._transition_queue_dict[0] = self._transition_queue
+
+    def add_model(self, model, initial=None):
+        super().add_model(model, initial)
+        if self.has_queue == 'model':
+            for mod in listify(model):
+                self._transition_queue_dict[id(mod) if mod != 'self' else id(self)] = deque()
+
     async def dispatch(self, trigger, *args, **kwargs):  # ToDo: not tested
         """ Trigger an event on all models assigned to the machine.
         Args:
@@ -352,14 +374,14 @@ class AsyncMachine(Machine):
         """
         return await asyncio.gather(*[func() for func in callables])
 
-    def switch_model_context(self, model):
+    async def switch_model_context(self, model):
         """
         This method is called by an `AsyncTransition` when all conditional tests have passed and the transition will happen.
         This requires already running tasks to be cancelled.
         Args:
             model (object): The currently processed model
         """
-        for running_task in self.async_tasks.get(model, []):
+        for running_task in self.async_tasks.get(id(model), []):
             if self.current_context.get() == running_task or running_task in self.protected_tasks:
                 continue
             elif running_task.done() is False:
@@ -379,23 +401,41 @@ class AsyncMachine(Machine):
         """
         if self.current_context.get() is None:
             self.current_context.set(asyncio.current_task())
-            if model in self.async_tasks:
-                self.async_tasks[model].append(asyncio.current_task())
+            if id(model) in self.async_tasks:
+                self.async_tasks[id(model)].append(asyncio.current_task())
             else:
-                self.async_tasks[model] = [asyncio.current_task()]
+                self.async_tasks[id(model)] = [asyncio.current_task()]
             try:
-                res = await self._process(func)
+                res = await self._process(func, model)
             except asyncio.CancelledError:
                 res = False
             finally:
-                self.async_tasks[model].remove(asyncio.current_task())
-                if len(self.async_tasks[model]) == 0:
-                    del self.async_tasks[model]
+                self.async_tasks[id(model)].remove(asyncio.current_task())
+                if len(self.async_tasks[id(model)]) == 0:
+                    del self.async_tasks[id(model)]
         else:
-            res = await self._process(func)
+            res = await self._process(func, model)
         return res
 
-    async def _process(self, trigger):
+    def remove_model(self, model):
+        """ Remove a model from the state machine. The model will still contain all previously added triggers
+        and callbacks, but will not receive updates when states or transitions are added to the Machine.
+        If an event queue is used, all queued events of that model will be removed."""
+        models = listify(model)
+        if self.has_queue == 'model':
+            for mod in models:
+                del self._transition_queue_dict[id(mod)]
+                self.models.remove(mod)
+        else:
+            for mod in models:
+                self.models.remove(mod)
+        if len(self._transition_queue) > 0:
+            queue = self._transition_queue
+            new_queue = [queue.popleft()] + [e for e in queue if e.args[0] not in models]
+            self._transition_queue.clear()
+            self._transition_queue.extend(new_queue)
+
+    async def _process(self, trigger, model):
         # default processing
         if not self.has_queue:
             if not self._transition_queue:
@@ -404,20 +444,22 @@ class AsyncMachine(Machine):
             else:
                 raise MachineError("Attempt to process events synchronously while transition queue is not empty!")
 
-        self._transition_queue.append(trigger)
+        self._transition_queue_dict[id(model)].append(trigger)
         # another entry in the queue implies a running transition; skip immediate execution
-        if len(self._transition_queue) > 1:
+        if len(self._transition_queue_dict[id(model)]) > 1:
             return True
 
-        # execute as long as transition queue is not empty ToDo: not tested!
-        while self._transition_queue:
+        while self._transition_queue_dict[id(model)]:
             try:
-                await self._transition_queue[0]()
-                self._transition_queue.popleft()
+                await self._transition_queue_dict[id(model)][0]()
             except Exception:
                 # if a transition raises an exception, clear queue and delegate exception handling
-                self._transition_queue.clear()
+                self._transition_queue_dict[id(model)].clear()
                 raise
+            try:
+                self._transition_queue_dict[id(model)].popleft()
+            except KeyError:
+                return True
         return True
 
 
@@ -450,7 +492,8 @@ class HierarchicalAsyncMachine(HierarchicalMachine, AsyncMachine):
 
     async def _trigger_event(self, _model, _trigger, _state_tree, *args, **kwargs):
         if _state_tree is None:
-            _state_tree = self._build_state_tree(listify(getattr(_model, self.model_attribute)), self.state_cls.separator)
+            _state_tree = self._build_state_tree(listify(getattr(_model, self.model_attribute)),
+                                                 self.state_cls.separator)
         res = {}
         for key, value in _state_tree.items():
             if value:
@@ -560,3 +603,15 @@ class AsyncTimeout(AsyncState):
     def on_timeout(self, value):
         """ Listifies passed values and assigns them to on_timeout."""
         self._on_timeout = listify(value)
+
+
+class _DictionaryMock(dict):
+
+    def __setitem__(self, key, item):
+        self._value = item
+
+    def __getitem__(self, key):
+        return self._value
+
+    def __repr__(self):
+        return repr("{{'*': {0}}}".format(self._value))
