@@ -91,27 +91,64 @@ class NestedEvent(Event):
     """
 
     def trigger(self, _model, _machine, *args, **kwargs):
-        """ Serially execute all transitions that match the current state,
-        halting as soon as one successfully completes. NOTE: This should only
+        """ Executes all transitions that match the current state,
+        halting as soon as one successfully completes. More precisely, it prepares a partial
+        of the internal ``_trigger`` function, passes this to ``machine._process``.
+        It is up to the machine's configuration of the Event whether processing happens queued (sequentially) or
+        whether further Events are processed as they occur. NOTE: This should only
         be called by HierarchicalMachine instances.
         Args:
-            _model (object): model object to
+            _model (object): The currently processed model
             _machine (HierarchicalMachine): Since NestedEvents can be used in multiple machine instances, this one
-                                            will be used to determine the current state separator.
+                                            will be used to determine the current state separator and the current scope.
             args and kwargs: Optional positional or named arguments that will
                 be passed onto the EventData object, enabling arbitrary state
                 information to be passed on to downstream triggered functions.
         Returns: boolean indicating whether or not a transition was
             successfully executed (True if successful, False if not).
         """
-        func = partial(self._trigger, _model, _machine, *args, **kwargs)
+        # Save the current scope (_machine.scoped, _machine.states, _machine.events) in partial
+        # since queued transitions could otherwise loose their scope.
+        func = partial(self._trigger, _model, _machine, (_machine.scoped, _machine.states, _machine.events),
+                       *args, **kwargs)
         # pylint: disable=protected-access
         # noinspection PyProtectedMember
         # Machine._process should not be called somewhere else. That's why it should not be exposed
         # to Machine users.
         return _machine._process(func)
 
-    def _trigger(self, _model, _machine, *args, **kwargs):
+    def _trigger(self, _model, _machine, _scope, *args, **kwargs):
+        """ Internal trigger function called by the ``HierarchicalMachine`` instance. This should not
+        be called directly but via the public method ``HierarchicalMachine.process``. In contrast to
+        the inherited ``Event._trigger``, this requires a scope tuple to process triggers in the right context.
+        Args:
+            _model (object): The currently processed model
+            _machine (HierarchicalMachine): The machine that should be used to process the event
+            _scope (Tuple): A tuple containing information about the currently scoped object, states an transitions.
+            args and kwargs: Optional positional or named arguments that will
+                be passed onto the EventData object, enabling arbitrary state
+                information to be passed on to downstream triggered functions.
+        Returns: boolean indicating whether or not a transition was
+            successfully executed (True if successful, False if not).
+        """
+        if _scope[0] != _machine.scoped:
+            with _machine(_scope):
+                return self._trigger_scoped(_model, _machine, *args, **kwargs)
+        else:
+            return self._trigger_scoped(_model, _machine, *args, **kwargs)
+
+    def _trigger_scoped(self, _model, _machine, *args, **kwargs):
+        """ Internal scope-adjusted trigger function called by the ``NestedEvent._trigger`` instance. This should not
+        be called directly.
+        Args:
+            _model (object): The currently processed model
+            _machine (HierarchicalMachine): The machine that should be used to process the event
+            args and kwargs: Optional positional or named arguments that will
+                be passed onto the EventData object, enabling arbitrary state
+                information to be passed on to downstream triggered functions.
+        Returns: boolean indicating whether or not a transition was
+            successfully executed (True if successful, False if not).
+        """
         state_tree = _machine._build_state_tree(getattr(_model, _machine.model_attribute), _machine.state_cls.separator)
         state_tree = reduce(dict.get, _machine.get_global_name(join=False), state_tree)
         ordered_states = _resolve_order(state_tree)
@@ -145,10 +182,19 @@ class NestedEvent(Event):
                     break
         except Exception as err:
             event_data.error = err
-            raise
+            if self.machine.on_exception:
+                self.machine.callbacks(self.machine.on_exception, event_data)
+            else:
+                raise
         finally:
-            machine.callbacks(machine.finalize_event, event_data)
-            _LOGGER.debug("%sExecuted machine finalize callbacks", machine.name)
+            try:
+                machine.callbacks(machine.finalize_event, event_data)
+                _LOGGER.debug("%sExecuted machine finalize callbacks", machine.name)
+            except Exception as err:
+                _LOGGER.error("%sWhile executing finalize callbacks a %s occurred: %s.",
+                              self.machine.name,
+                              type(err).__name__,
+                              str(err))
         return event_data.result
 
 
@@ -157,7 +203,7 @@ class NestedState(State):
     Attributes:
         states (OrderedDict): A list of substates of the current state.
         events (dict): A list of events defined for the nested state.
-        initial (str): Name of a child which should be entered when the state is entered.
+        initial (list, str, NestedState or Enum): (Name of a) child or list of children that should be entered when the state is entered.
         exit_stack (defaultdict): A list of currently active substates
     """
 
@@ -192,13 +238,17 @@ class NestedState(State):
 
     def scoped_enter(self, event_data, scope=[]):
         self._scope = scope
-        self.enter(event_data)
-        self._scope = []
+        try:
+            self.enter(event_data)
+        finally:
+            self._scope = []
 
     def scoped_exit(self, event_data, scope=[]):
         self._scope = scope
-        self.exit(event_data)
-        self._scope = []
+        try:
+            self.exit(event_data)
+        finally:
+            self._scope = []
 
     @property
     def name(self):
@@ -289,7 +339,8 @@ class NestedTransition(Transition):
                     scoped_tree[state.name] = OrderedDict()
                     if state.initial:
                         q.append((scoped_tree[state.name], prefix + [state.name],
-                                  [state.states[i] for i in listify(state.initial)]))
+                                  [state.states[i.name] if hasattr(i, 'name') else state.states[i]
+                                   for i in listify(state.initial)]))
                 if not q:
                     break
                 scoped_tree, prefix, initial_states = q.pop(0)
@@ -322,7 +373,8 @@ class NestedTransition(Transition):
 
 class HierarchicalMachine(Machine):
     """ Extends transitions.core.Machine by capabilities to handle nested states.
-        A hierarchical machine REQUIRES NestedStates (or any subclass of it) to operate.
+        A hierarchical machine REQUIRES NestedStates, NestedEvent and NestedTransitions
+        (or any subclass of it) to operate.
     """
 
     state_cls = NestedState
@@ -330,6 +382,9 @@ class HierarchicalMachine(Machine):
     event_cls = NestedEvent
 
     def __init__(self, *args, **kwargs):
+        assert issubclass(self.state_cls, NestedState)
+        assert issubclass(self.event_cls, NestedEvent)
+        assert issubclass(self.transition_cls, NestedTransition)
         self._stack = []
         self.scoped = self
         _super(HierarchicalMachine, self).__init__(*args, **kwargs)
@@ -407,9 +462,10 @@ class HierarchicalMachine(Machine):
     def add_states(self, states, on_enter=None, on_exit=None, ignore_invalid_triggers=None, **kwargs):
         """ Add new nested state(s).
         Args:
-            states (list, str, dict, Enum or NestedState): a list, a NestedState instance, the
+            states (list, str, dict, Enum, NestedState or Machine): a list, a NestedState instance, the
                 name of a new state, an enumeration (member) or a dict with keywords to pass on to the
-                NestedState initializer. If a list, each element can be a string, NestedState or enumeration member.
+                NestedState initializer. If a list, each element can be a string, dict, NestedState or
+                enumeration member.
             on_enter (str or list): callbacks to trigger when the state is
                 entered. Only valid if first argument is string.
             on_exit (str or list): callbacks to trigger when the state is
@@ -452,6 +508,10 @@ class HierarchicalMachine(Machine):
                     self.states[new_state.name] = new_state
                     self._init_state(new_state)
             elif isinstance(state, Enum):
+                if self.state_cls.separator in state.name:
+                    raise ValueError("State '{0}' contains '{1}' which is used as state name separator. "
+                                     "Consider changing the NestedState.separator to avoid this issue."
+                                     "".format(state.name, self.state_cls.separator))
                 if remap is not None and state.name in remap:
                     return
                 new_state = self._create_state(state, on_enter=on_enter, on_exit=on_exit,
@@ -467,19 +527,22 @@ class HierarchicalMachine(Machine):
                 remap = state.pop('remap', None)
                 if 'ignore_invalid_triggers' not in state:
                     state['ignore_invalid_triggers'] = ignore
-                state_children = state.pop('children', [])
+
+                # parallel: [states] is just a short handle for {children: [states], initial: [state_names]}
                 state_parallel = state.pop('parallel', [])
+                if state_parallel:
+                    state_children = state_parallel
+                    state['initial'] = [s['name'] if isinstance(s, dict)
+                                        else s for s in state_children]
+                else:
+                    state_children = state.pop('children', state.pop('states', []))
                 transitions = state.pop('transitions', [])
                 new_state = self._create_state(**state)
                 self.states[new_state.name] = new_state
                 self._init_state(new_state)
                 remapped_transitions = []
                 with self(new_state.name):
-                    if state_parallel:
-                        self.add_states(state_parallel, remap=remap, **kwargs)
-                        new_state.initial = [s if isinstance(s, string_types) else s['name'] for s in state_parallel]
-                    else:
-                        self.add_states(state_children, remap=remap, **kwargs)
+                    self.add_states(state_children, remap=remap, **kwargs)
                     if remap is not None:
                         drop_event = []
                         for evt in self.events.values():
@@ -549,7 +612,8 @@ class HierarchicalMachine(Machine):
         _super(HierarchicalMachine, self).add_transition(trigger, source, dest, conditions,
                                                          unless, before, after, prepare, **kwargs)
 
-    def get_global_name(self, state=None, join=True):
+    def get_global_name(self, state=None, join=True, scope=None):
+        scope = scope or self
         local_stack = [s[0] for s in self._stack] + [self.scoped]
         local_stack_start = len(local_stack) - local_stack[::-1].index(self)
         domains = [s.name for s in local_stack[local_stack_start:]]
@@ -655,24 +719,41 @@ class HierarchicalMachine(Machine):
                 res.append(self.get_state(state))
         return res
 
-    def get_transitions(self, trigger="", source="*", dest="*"):
+    def get_transitions(self, trigger="", source="*", dest="*", delegate=False):
+        """ Return the transitions from the Machine.
+        Args:
+            trigger (str): Trigger name of the transition.
+            source (str, State or Enum): Limits list to transitions from a certain state.
+            dest (str, State or Enum): Limits list to transitions to a certain state.
+            delegate (Optional[bool]): If True, consider delegations to parents of source
+        """
         with self():
-            source_path = [] if source == "*" else source.split(self.state_cls.separator)
-            dest_path = [] if dest == "*" else dest.split(self.state_cls.separator)
-            if source_path:
-                transitions = []
-                while source_path:
-                    transitions.extend(self.get_nested_transitions(trigger, source_path, dest_path))
-                    source_path.pop()
-                    return transitions
-            else:
-                return self.get_nested_transitions(trigger, None, dest_path)
+            source_path = [] if source == "*" \
+                else source.split(self.state_cls.separator) if isinstance(source, string_types) \
+                else self._get_enum_path(source) if isinstance(source, Enum) \
+                else self._get_state_path(source)
+            dest_path = [] if dest == "*" \
+                else dest.split(self.state_cls.separator) if isinstance(dest, string_types) \
+                else self._get_enum_path(dest) if isinstance(dest, Enum) \
+                else self._get_state_path(dest)
+            matches = self.get_nested_transitions(trigger, source_path, dest_path)
+            # only consider delegations when source_path contains a nested state (len > 1)
+            if delegate is False or len(source_path) < 2:
+                return matches
+            source_path.pop()
+            while source_path:
+                matches.extend(self.get_transitions(trigger,
+                                                    source=self.state_cls.separator.join(source_path),
+                                                    dest=dest))
+                source_path.pop()
+            return matches
 
     def get_triggers(self, *args):
         """ Extends transitions.core.Machine.get_triggers to also include parent state triggers. """
         triggers = []
         with self():
-            for state_name in args:
+            for state in args:
+                state_name = state.name if hasattr(state, 'name') else state
                 state_path = state_name.split(self.state_cls.separator)
                 if len(state_path) > 1:  # we only need to check substates when 'state_name' refers to a substate
                     with self(state_path[0]):
@@ -682,6 +763,18 @@ class HierarchicalMachine(Machine):
                     state_path.pop()
         return triggers
 
+    def has_trigger(self, trigger, state=None):
+        """ Check whether an event/trigger is known to the machine
+        Args:
+            trigger (str): Event/trigger name
+            state (optional[NestedState]): Limits the recursive search to this state and its children
+        Returns:
+            bool: True if event is known and False otherwise
+        """
+
+        state = state or self
+        return trigger in state.events or any([self.has_trigger(trigger, sta) for sta in state.states.values()])
+
     def is_state(self, state_name, model, allow_substates=False):
         current_name = getattr(model, self.model_attribute)
         if allow_substates:
@@ -763,7 +856,7 @@ class HierarchicalMachine(Machine):
 
     def _add_model_to_state(self, state, model):
         name = self.get_global_name(state)
-        if self.state_cls.separator == '_' or self.state_cls.separator not in name:
+        if self.state_cls.separator == '_':
             value = state.value if isinstance(state.value, Enum) else name
             self._checked_assignment(model, 'is_%s' % name, partial(self.is_state, value, model))
             # Add dynamic method callbacks (enter/exit) if there are existing bound methods in the model
@@ -773,6 +866,14 @@ class HierarchicalMachine(Machine):
                 if hasattr(model, method) and inspect.ismethod(getattr(model, method)) and \
                         method not in getattr(state, callback):
                     state.add_callback(callback[3:], method)
+        else:
+            path = name.split(self.state_cls.separator)
+            value = state.value if isinstance(state.value, Enum) else name
+            trig_func = partial(self.is_state, value, model)
+            if hasattr(model, 'is_' + path[0]):
+                getattr(model, 'is_' + path[0]).add(trig_func, path[1:])
+            else:
+                self._checked_assignment(model, 'is_' + path[0], FunctionWrapper(trig_func, path[1:]))
         with self(state.name):
             for event in self.events.values():
                 if not hasattr(model, event.name):
@@ -821,6 +922,16 @@ class HierarchicalMachine(Machine):
                     return res
         return []
 
+    def _get_state_path(self, state, prefix=[]):
+        if state in self.states.values():
+            return prefix + [state.name]
+        for name in self.states:
+            with self(name):
+                res = self._get_state_path(state, prefix=prefix + [name])
+                if res:
+                    return res
+        return []
+
     def _check_event_result(self, res, model, trigger):
         if res is None:
             state_names = getattr(model, self.model_attribute)
@@ -830,7 +941,12 @@ class HierarchicalMachine(Machine):
                 ignore = state.ignore_invalid_triggers if state.ignore_invalid_triggers is not None \
                     else self.ignore_invalid_triggers
                 if not ignore:
-                    raise MachineError(msg)
+                    # determine whether a MachineError (valid event but invalid state) ...
+                    if self.has_trigger(trigger):
+                        raise MachineError(msg)
+                    # or AttributeError (invalid event) is appropriate
+                    else:
+                        raise AttributeError("Do not know event named '%s'." % trigger)
             _LOGGER.warning(msg)
             res = False
         return res
@@ -845,10 +961,7 @@ class HierarchicalMachine(Machine):
         Returns:
             bool: True if a transitions has been conducted or the trigger event has been queued.
         """
-        try:
-            return self.trigger_event(model, trigger_name, *args, **kwargs)
-        except MachineError:
-            raise AttributeError("Do not know event named '%s'." % trigger_name)
+        return self.trigger_event(model, trigger_name, *args, **kwargs)
 
     def _has_state(self, state, raise_error=False):
         """ This function
@@ -937,7 +1050,7 @@ class HierarchicalMachine(Machine):
                     tmp = self._trigger_event(_model, _trigger, value, *args, **kwargs)
                     if tmp is not None:
                         res[key] = tmp
-            if not res.get(key, None) and _trigger in self.events:
+            if res.get(key, False) is False and _trigger in self.events:
                 tmp = self.events[_trigger].trigger(_model, self, *args, **kwargs)
                 if tmp is not None:
                     res[key] = tmp
